/*****************************************************************************
 * Copyright 2019 Johann 'Myrkraverk' Oskarsson                              *
 * Email: johann@myrkraverk.com                                              *
 * License: WTFPL 2.0                                                        *
 * Version: 1.0                                                              *
 *                                                                           *
 * A very basic example for using NSS to calculate a SHA256 sum.             *
 *****************************************************************************/

/* Include the NSPR headers.
 */
#include <nspr/prio.h>
#include <nspr/prprf.h>

/* Include the NSS headers.
 */
#include <nss/nss.h>
#include <nss/pk11pub.h>

/* This program calculates the SHA256 of the string "abc" as a
 * demonstration of how to use NSS to calculate a hash.
 *
 * This code uses PR_fprintf() exclusively for printing messages.  The
 * first argument is the struct PrFileDesc * pointer where we want the
 * message to output, and PR_STDOUT and PR_STDERR are macros that give
 * us standard output and standard error pointers, respectively.  Next
 * is a format string, followed by arguments according to the format
 * string, just like printf().
 *
 * Notice that our success messages are printed to standard output,
 * while our error messages are printed to standard error.
 *
 * As a demonstration of actual production code, error checking is in
 * place for all cryptographic functions.  However, we're not totally
 * pedantic and don't check the return value of PR_fprintf() which can
 * actually fail if, say, standard output is piped to a file and
 * there's no space left on the file system.  This is worth fixing for
 * an actualy production system.
 */
int main( int argc, char *argv[] ) {

  /* Return value from main(), failure by default.  This is arguably
   * better style than it is to succeed by default.  When all the
   * steps have been completed successfully, we can set this to zero.
   */
  int ret = 1;

  /* Initialise the NSS library, and check for the return value.
   */
  SECStatus s = NSS_NoDB_Init(".");
  if (s != SECSuccess) {
    /* If the NSS libraly fails to initialise, we print a short
     * message and return immediately.  The shutdown code below is for
     * when we've done at least this step successfully.
    */
    PR_fprintf( PR_STDERR, "NSS initialisation failed!\n" );
    return ret;
  } else {
    /* And on success, we print a nice message, but in production code
     * we normally wouldn't do this.
     */
    PR_fprintf( PR_STDOUT, "NSS initialisation successful.\n" );
  }
  
  /* We explicitly create a SHA256 hashing context.
   */
  PK11Context *context = PK11_CreateDigestContext( SEC_OID_SHA256 );
  /* If the context creation fails, we print a short error message and
     jump to the shutdown code below.
   */
  if ( !context ) {
    PR_fprintf( PR_STDERR, "Unable to create digest context.\n" );
    goto shutdown;
  }

  /* We initialise the context before we begin to calculate the hash.
   */
  s = PK11_DigestBegin( context );
  /* If the context initialisation fails, we print a short error
   * message and jump to the shutdown code below.
   */
  if ( s != SECSuccess ) {
    PR_fprintf( PR_STDERR, "Error starting the digest context.\n" );
    goto shutdown;
  }

  /* Our test vector for the SHA256 hashing algorithm.  Doing it like
   * this, with an explicit array initialiser means there will be no
   * zero byte at the end of the string.
   */
  unsigned char vector[] = { 'a', 'b', 'c' }; 

  /* calculate the digest of the vector; `context' is the context of
   * course, `vector' is the buffer we want to hash, and `sizeof
   * vector' is the length of the buffer.
   */
  s = PK11_DigestOp( context, vector, sizeof vector );
  /* If the hashing operation fails, we print a short error message
   * and jump to the shutdown code below.
   */
  if ( s != SECSuccess ) {
    PR_fprintf( PR_STDERR, "Unable to hash test vector.\n" );
    goto shutdown;
  }

  /* The digest itself is 256 / 8 = 32 bytes long for SHA256.  We let
   * the computer calculate it.  The digest and length are unsigned to
   * fit the parameters of the PK11_DigestFinal() function.  The first
   * parameter is the context, of course, 'digest' is the output
   * buffer, the `&length' parameter is where the actual length of the
   * digest is stored, while the fourth parameter, `sizeof digest' is
   * the length of the `digest' buffer.
   */
  unsigned char digest[ 256 / 8 ];
  unsigned int length = 0;
  s = PK11_DigestFinal( context, digest, &length, sizeof digest );
  /* If the digest finalisation fails, we print a short error message
   * and jump to the shutdown code below.
   */
  if ( s != SECSuccess ) {
    PR_fprintf( PR_STDERR, "Unable to finalise the hash.\n" );
    goto shutdown;
  }

  /* We print out how many bytes we calculated for the SHA256 hash.
   */
  PR_fprintf( PR_STDOUT, "Hashed into %d bytes.\n", length );

  /* Then print the hexadecmial value of the digest.  PR_fprintf()
   * doesn't have %hhx for 8bit values, so we fake it with an integer.
   *
   * Here we use an unsigned `i' because `length' is unsigned.
   */
  for( unsigned int i = 0; i < length; i++ ) {
    PR_fprintf( PR_STDOUT, "%02x", ( int ) digest[ i ] );
  }
  /* And finally print a newline.
   */
  PR_fprintf( PR_STDOUT, "\n" );

  /* Now at the end of the program, we allow ourselves to return
   * success to the operating system.
   */
  ret = 0;

 shutdown:
  /* We destroy and free the context; `context' is the context, and
   * `PR_TRUE' means we'll also free it.  And even though we're at the
   * end of the function, we still set the pointer to NULL for safety.
   */
  if ( context ) {
    PK11_DestroyContext( context, PR_TRUE );
    context = NULL;
  }
  /* We shut down the NSS library.  This can actually fail, and will
   * do so if there are pending digest contexts that haven't been
   * destroyed yet.
   *
   * We set the return value to fail because it could be success by
   * now, and we use a different value from the other error messages.
   */
  if ( NSS_Shutdown() != SECSuccess ) {
    PR_fprintf( PR_STDERR, "NSS Shutdown failed!\n" );
    ret = 2;
  }
  
  return ret;
}
